@@ -6,8 +6,9 @@ use Carp;
 package Template::Declare;
 use Template::Declare::Buffer;
 use Class::ISA;
+use String::BufferStack;
 
-our $VERSION = "0.29";
+our $VERSION = "0.35";
 
 use base 'Class::Data::Inheritable';
 __PACKAGE__->mk_classdata('roots');
@@ -16,7 +17,7 @@ __PACKAGE__->mk_classdata('aliases');
 __PACKAGE__->mk_classdata('alias_metadata');
 __PACKAGE__->mk_classdata('templates');
 __PACKAGE__->mk_classdata('private_templates');
-__PACKAGE__->mk_classdata('buffer_stack');
+__PACKAGE__->mk_classdata('buffer');
 __PACKAGE__->mk_classdata('imported_into');
 __PACKAGE__->mk_classdata('around_template');
 
@@ -26,10 +27,10 @@ __PACKAGE__->aliases(           {} );
 __PACKAGE__->alias_metadata(    {} );
 __PACKAGE__->templates(         {} );
 __PACKAGE__->private_templates( {} );
-__PACKAGE__->buffer_stack( [] );
+__PACKAGE__->buffer( String::BufferStack->new );
 __PACKAGE__->around_template( undef );
 
-__PACKAGE__->new_buffer_frame();
+*String::BufferStack::data = \&String::BufferStack::buffer;
 
 use vars qw/$TEMPLATE_VARS/;
 
@@ -351,20 +352,11 @@ sub init {
 }
 
 sub new_buffer_frame {
-    my $buffer = Template::Declare::Buffer->new();
-    unshift @{ __PACKAGE__->buffer_stack }, $buffer;
-
+    __PACKAGE__->buffer->push( private => 1 );
 }
 
 sub end_buffer_frame {
-    shift @{ __PACKAGE__->buffer_stack };
-}
-
-sub buffer {
-    unless ( __PACKAGE__->buffer_stack->[0] ) {
-        Carp::confess( __PACKAGE__ . "->buffer called with no buffer" );
-    }
-    return __PACKAGE__->buffer_stack->[0];
+    __PACKAGE__->buffer->pop;
 }
 
 =head2 show TEMPLATE_NAME
@@ -740,6 +732,26 @@ Note you can always get rid of the C<outs> crap if the string literal is the onl
 
    p { 'hello, world!' }
 
+=item *
+
+Look out! If the if block is the last block/statement and the condition part is evaluated to be 0:
+
+   p { if ( 0 ) { } }
+
+produces
+
+   <p>0</p>
+
+instead of the more intutive output:
+
+   <p></p>
+
+This's because 0 is the last expression, so it's returned as the value of the whole block, which is used as the content of <p> tag.
+
+To get rid of this, just put an empty string at the end so it returns empty string as the content instead of 0:
+
+   p { if ( 0 ) { } '' }
+
 =back
 
 =head1 BUGS
@@ -766,9 +778,11 @@ L<Template::Declare::Tags>, L<Template::Declare::TagSet>, L<Template::Declare::T
 
 Jesse Vincent <jesse@bestpractical.com>
 
-=head1 COPYRIGHT
+=head1 LICENSE
 
-Copyright 2006-2007 Best Practical Solutions, LLC
+Template::Declare is Copyright 2006-2008 Best Practical Solutions, LLC.
+
+Template::Declare is distributed under the same terms as Perl itself.
 
 =cut
 
