@@ -418,49 +418,30 @@ sub xml_decl (&;$) {
 
 C<outs> HTML-encodes its arguments and appends them to C<Template::Declare>'s output buffer.
 
-
-=cut
-
-#sub outs { outs_raw( map { _postprocess($_); } grep {defined} @_ ); }
-
 =head2 outs_raw STUFF
 
 C<outs_raw> appends its arguments to C<Template::Declare>'s output buffer without doing any HTML escaping.
 
 =cut
 
-#sub outs_raw { Template::Declare->buffer->append( join( '', grep {defined} @_ )); return ''; }
-
 sub outs     { _outs( 0, @_ ); }
 sub outs_raw { _outs( 1, @_ ); }
 
 sub _outs {
     my $raw     = shift;
     my @phrases = (@_);
-    my $buf;
-    Template::Declare->new_buffer_frame;
+
+    Template::Declare->buffer->push( private => (defined wantarray and not wantarray) );
 
     foreach my $item ( grep {defined} @phrases ) {
-
-        Template::Declare->new_buffer_frame;
-        my $returned =
-            ref($item) eq 'CODE'
+        my $returned = ref($item) eq 'CODE'
             ? $item->()
-            : ( $raw ? $item : _postprocess($item) ) || '';
-        my $content = Template::Declare->buffer->data || '';
-        Template::Declare->end_buffer_frame;
-        Template::Declare->buffer->append( $content . $returned );
+            : $raw
+                ? $item
+                : _postprocess($item);
+        Template::Declare->buffer->append( $returned );
     }
-
-    $buf = Template::Declare->buffer->data || '';
-    Template::Declare->end_buffer_frame;
-    if ( defined wantarray and not wantarray ) {
-        return $buf;
-    } else {
-        Template::Declare->buffer->append($buf);
-
-    }
-    return '';
+    return Template::Declare->buffer->pop;
 }
 
 =head2 get_current_attr
@@ -601,9 +582,9 @@ care of putting the output in the right place and tidying up after itself.
 
 sub smart_tag_wrapper (&) {
     my $coderef = shift;
-    my $buf     = "\n";
 
-    Template::Declare->new_buffer_frame;
+    Template::Declare->buffer->append("\n");
+    Template::Declare->buffer->push;
 
     my %attr = %ATTRIBUTES;
     %ATTRIBUTES = ();                              # prevent leakage
@@ -612,18 +593,9 @@ sub smart_tag_wrapper (&) {
         map { ref($_) ? $_ : _postprocess($_) }    #
         $coderef->(%attr);
 
-
-    if ( length( Template::Declare->buffer->data ) ) {
-
-        # We concatenate to force scalarization when $last or
-        # $Template::Declare->buffer is solely a Jifty::Web::Link
-        $buf .= Template::Declare->buffer->data;
-    } elsif ( length $last ) {
-        $buf .= $last;
-    }
-
-    Template::Declare->end_buffer_frame;
-    Template::Declare->buffer->append($buf);
+    my $has_content = Template::Declare->buffer->length;
+    Template::Declare->buffer->pop;
+    Template::Declare->buffer->append( "$last" ) if not $has_content and length $last;
 
     return '';
 }
@@ -645,13 +617,15 @@ sub _tag {
     $tag = $tagset->namespace . ":$tag"
         if defined $tagset->namespace;
 
-    my $buf = "\n" . ( " " x $TAG_NEST_DEPTH ) . "<$tag"
-        . join( '',
-        map { qq{ $_="} . ( $ATTRIBUTES{$_} || '' ) . qq{"} }
-            sort keys %ATTRIBUTES );
-
-    my $had_content = 0;
+    Template::Declare->buffer->append(
+              "\n" 
+            . ( " " x $TAG_NEST_DEPTH ) . "<$tag"
+            . join( '',
+            map { qq{ $_="} . ( $ATTRIBUTES{$_} || '' ) . qq{"} }
+                sort keys %ATTRIBUTES )
+    );
 
+    my $attrs = "";
     {
         no warnings qw( uninitialized redefine once );
 
@@ -674,42 +648,30 @@ sub _tag {
             my $field = shift;
             my $val   = shift;
 
-            $buf .= ' ' . $field . q{="} . _postprocess($val, 1) . q{"};
+            $attrs .= ' ' . $field . q{="} . _postprocess($val, 1) . q{"};
             wantarray ? () : '';
         };
 
         local $TAG_NEST_DEPTH = $TAG_NEST_DEPTH + 1;
         %ATTRIBUTES = ();
-        Template::Declare->new_buffer_frame;
-        my $last = join '', map { ref($_) ? $_ : _postprocess($_) } $code->();
-
-        if ( length( Template::Declare->buffer->data ) ) {
-
-# We concatenate to force scalarization when $last or $Template::Declare->buffer is solely a Jifty::Web::Link
-            $buf .= '>' . Template::Declare->buffer->data;
-            $had_content = 1;
-        } elsif ( length $last ) {
-            $buf .= '>' . $last;
-            $had_content = 1;
-        } else {
-            $had_content = 0;
-        }
-
-        Template::Declare->end_buffer_frame;
-
+        Template::Declare->buffer->push( private => 1 );
+        my $last = join '', map { ref($_) && $_->isa('Template::Declare::Tag') ? $_ : _postprocess($_) } $code->();
+        Template::Declare->buffer->append("$last") if not Template::Declare->buffer->length and length $last;
     }
+    my $content = Template::Declare->buffer->pop;
+    Template::Declare->buffer->append($attrs);
 
-    if ($had_content) {
-        $buf .= "\n" . ( " " x $TAG_NEST_DEPTH ) if ( $buf =~ /\>$/ );
-        $buf .= "</$tag>";
+    if (length $content) {
+        Template::Declare->buffer->append(">$content");
+        Template::Declare->buffer->append("\n" . ( " " x $TAG_NEST_DEPTH )) if $content =~ /\</;
+        Template::Declare->buffer->append("</$tag>");
     } elsif ( $tagset->can_combine_empty_tags($tag) ) {
-        $buf .= " />";
+        Template::Declare->buffer->append(" />");
     } else {
         # Otherwise we supply a closing tag.
-        $buf .= "></$tag>";
+        Template::Declare->buffer->append("></$tag>");
     }
 
-    Template::Declare->buffer->append($buf);
     return ( ref($more_code) && $more_code->isa('CODE') )
         ? $more_code->()
         : '';
@@ -738,64 +700,57 @@ C<Template::Declare> package.
 
 sub show {
     my $template = shift;
-    my $args  = \@_;
-    my $data;
 
     # if we're inside a template, we should show private templates
     if ( caller->isa('Template::Declare') ) {
-       _show_template( $template, 1, $args );
+        _show_template( $template, 1, \@_ );
         return Template::Declare->buffer->data;
     } else {
-        show_page( $template, $args);
+        show_page( $template, @_);
     }
 
 }
 
-
-
 sub show_page {
-    my $template        = shift;
+    my $template = shift;
     my $args = \@_;
-    my $INSIDE_TEMPLATE = 0;
 
-    # if we're inside a template, we should show private templates
-    Template::Declare->new_buffer_frame;
-    _show_template( $template, 0, $args );
-    my $data = Template::Declare->buffer->data;
-    Template::Declare->end_buffer_frame;
-    %ELEMENT_ID_CACHE = ();    # We're done. we can clear the cache
-    if (not defined wantarray()) {
-        Template::Declare->buffer->append($data);
+    if (defined wantarray) {
+        Template::Declare->buffer->push( private => 1 );
+        _show_template( $template, 0, $args );
+        %ELEMENT_ID_CACHE = ();
+        return Template::Declare->buffer->pop;
+    } else {
+        _show_template( $template, 0, $args );
+        %ELEMENT_ID_CACHE = ();
         return undef;
-     } else {
-        return $data;
-     }
+    }
 }
 
-sub _resolve_relative_template_path {
+sub _resolve_template_path {
     my $template = shift;
 
-    return $template if ( $template =~ '^\/' );
-    my $parent = current_template();
+    my @parts;
+    if ( substr($template, 0, 1) ne '/' ) {
+        # relative
+        @parts = split '/', current_template();
+        # Get rid of the parent's template name
+        pop @parts;
+    }
 
-    my @parent   = split( '/', $parent );
-    my @template = split( '/', $template );
-
-    @template = grep { $_ !~ /^\.$/} @template; # Get rid of "." entries
-
-    # Let's find out how many levels they want to pop up
-    my @uplevels = grep { /^\.\.$/ } @template;
-    @template = grep { $_ !~ /^\.\.$/ } @template;
-
-
-
-    pop @parent;            # Get rid of the parent's template name
-    pop @parent for @uplevels;
-    return (join( '/', @parent, @template ) );
+    foreach ( split '/', $template ) {
+        if ( $_ eq '..' ) {
+            pop @parts;
+        }
+        # Get rid of "." and empty entries by the way
+        elsif ( $_ ne '.' && $_ ne '' ) {
+            push @parts, $_;
+        }
+    }
 
+    return join '/', @parts;
 }
 
-
 =head2 current_template
 
 Returns the absolute path of the current template
@@ -811,9 +766,8 @@ sub _show_template {
     my $template        = shift;
     my $inside_template = shift;
     my $args = shift;
-    local @TEMPLATE_STACK  = @TEMPLATE_STACK;
-    $template = _resolve_relative_template_path($template);
-    push @TEMPLATE_STACK, $template;
+    $template = _resolve_template_path($template);
+    local @TEMPLATE_STACK  = (@TEMPLATE_STACK, $template);
 
     my $callable =
         ( ref($template) && $template->isa('Template::Declare::Tag') )
@@ -847,6 +801,10 @@ sub _postprocess {
     my $val = shift;
     my $skip_postprocess = shift;
 
+    return $val unless defined $val;
+
+    # stringify in case $val is object with overloaded ""
+    $val = "$val";
     if ( ! $SKIP_XML_ESCAPING ) {
         no warnings 'uninitialized';
         $val =~ s/&/&#38;/g;
@@ -858,7 +816,7 @@ sub _postprocess {
         $val =~ s/'/&#39;/g;
     }
     $val = Template::Declare->postprocessor->($val)
-        if defined($val) && !$skip_postprocess;
+        unless $skip_postprocess;
 
     return $val;
 }
@@ -951,13 +909,10 @@ sub stringify {
     my $self = shift;
 
     if ( defined wantarray ) {
-        Template::Declare->new_buffer_frame;
+        Template::Declare->buffer->push( private => 1 );
         my $returned = $self->();
-        my $content  = Template::Declare->buffer->data();
-        Template::Declare->end_buffer_frame;
-        return ( $content . $returned );
+        return Template::Declare->buffer->pop . $returned;
     } else {
-
         return $self->();
     }
 }
